"""Mass spectrometry file type input/output operations."""
import collections
import csv
import os

from .. import __version__


class MztabWriter:
    """
    Export spectrum identifications to an mzTab file.

    Parameters
    ----------
    filename : str
        The name of the mzTab file.
    """

    def __init__(self, filename: str):
        self.filename = filename

        self.psms = []

    def save(self) -> None:
        """
        Export the spectrum identifications to the mzTab file.
        """
        with open(self.filename, "w") as f:
            writer = csv.writer(f, delimiter="\t", lineterminator=os.linesep)

            writer.writerow(
                [
                    "PSH",
                    "sequence",
                    "PSM_ID",
                    "accession",
                    "unique",
                    "database",
                    "database_version",
                    "search_engine",
                    "search_engine_score[1]",
                    "modifications",
                    "retention_time",
                    "charge",
                    "exp_mass_to_charge",
                    "calc_mass_to_charge",
                    "spectra_ref",
                    "pre",
                    "post",
                    "start",
                    "end",
                    "opt_ms_run[1]_aa_scores",
                ]
            )
            for psm in self.psms:
                writer.writerow(
                    [
                        "PSM",
                        psm[0],  # sequence
                        str(psm[1]),  # PSM_ID
                        "null",  # accession
                        "null",  # unique
                        "null",  # database
                        "null",  # database_version
                        f"[MS, MS, DiffNovo_DIA, {__version__}]",
                        psm[2],  # search_engine_score[1]
                        # FIXME: Modifications should be specified as
                        #  controlled vocabulary terms.
                        "null",  # modifications
                        # FIXME: Can we get the retention time from the data
                        #  loader?
                        "null",  # retention_time
                        psm[3],  # charge
                        psm[4],  # exp_mass_to_charge
                        psm[5],  # calc_mass_to_charge
                        f"ms_run[1]:index={psm[1]}",  # spectra_ref
                        "null",  # pre
                        "null",  # post
                        "null",  # start
                        "null",  # end
                        psm[6],  # opt_ms_run[1]_aa_scores
                    ]
                )



"""The Pytorch Datasets"""
from . import preprocessing

from .datasets import (
    SpectrumDataset,
    AnnotatedSpectrumDataset,
)

from .hdf5 import SpectrumIndex, AnnotatedSpectrumIndex

from .loaders import SpectrumDataModule



"""A de novo peptide sequencing model"""
import logging

import torch
import einops
import numpy as np
import pytorch_lightning as pl

from ...components import SpectrumEncoder, PeptideDecoder, ModelMixin

LOGGER = logging.getLogger(__name__)


class Spec2Pep(pl.LightningModule, ModelMixin):
    """A Transformer model for de novo peptide sequencing.

    Use this model in conjunction with a pytorch-lightning Trainer.

    Parameters
    ----------
    dim_model : int, optional
        The latent dimensionality used by the Transformer model.
    n_head : int, optional
        The number of attention heads in each layer. ``dim_model`` must be
        divisible by ``n_head``.
    dim_feedforward : int, optional
        The dimensionality of the fully connected layers in the Transformer
        model.
    n_layers : int, optional
        The number of Transformer layers.
    dropout : float, optional
        The dropout probability for all layers.
    dim_intensity : int or None, optional
        The number of features to use for encoding peak intensity.
        The remaining (``dim_model - dim_intensity``) are reserved for
        encoding the m/z value. If ``None``, the intensity will be projected
        up to ``dim_model`` using a linear layer, then summed with the m/z
        emcoding for each peak.
    custom_encoder : SpectrumEncoder or PairedSpectrumEncoder, optional
        A pretrained encoder to use. The ``dim_model`` of the encoder must
        be the same as that specified by the ``dim_model`` parameter here.
    max_length : int, optional
        The maximum peptide length to decode.
    residues: Dict or str {"massivekb", "canonical"}, optional
        The amino acid dictionary and their masses. By default this is only
        the 20 canonical amino acids, with cysteine carbamidomethylated. If
        "massivekb", this dictionary will include the modifications found in
        MassIVE-KB. Additionally, a dictionary can be used to specify a custom
        collection of amino acids and masses.
    max_charge : int, optional
        The maximum charge state to consider.
    beam_size : int, optional
        The number of paths to pursue when decoding.
    n_log : int, optional
        The number of epochs to wait between logging messages.
    **kwargs : Dict
        Keyword arguments passed to the Adam optimizer
    """

    def __init__(
        self,
        dim_model=128,
        n_head=8,
        dim_feedforward=1024,
        n_layers=1,
        dropout=0,
        dim_intensity=None,
        custom_encoder=None,
        max_length=100,
        residues="canonical",
        max_charge=5,
        beam_size=3,
        n_log=10,
        **kwargs,
    ):
        """Initialize a Spec2Pep model"""
        super().__init__()

        # Writable
        self.max_length = max_length
        self.n_log = n_log
        self.beam_size = beam_size

        # Build the model
        if custom_encoder is not None:
            self.encoder = custom_encoder
        else:
            self.encoder = SpectrumEncoder(
                dim_model=dim_model,
                n_head=n_head,
                dim_feedforward=dim_feedforward,
                n_layers=n_layers,
                dropout=dropout,
                dim_intensity=dim_intensity,
            )

        self.decoder = PeptideDecoder(
            dim_model=dim_model,
            n_head=n_head,
            dim_feedforward=dim_feedforward,
            n_layers=n_layers,
            dropout=dropout,
            residues=residues,
            max_charge=max_charge,
        )

        self.softmax = torch.nn.Softmax(dim=2)
        self.celoss = torch.nn.CrossEntropyLoss(ignore_index=0)

        # Things for training
        self._history = []
        self.opt_kwargs = kwargs
        self.stop_token = self.decoder._aa2idx["$"]

    def forward(self, spectra, precursors):
        """Sequence a batch of mass spectra.

        Parameters
        ----------
        spectrum : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra to embed. Axis 0 represents a mass spectrum, axis 1
            contains the peaks in the mass spectrum, and axis 2 is essentially
            a 2-tuple specifying the m/z-intensity pair for each peak. These
            should be zero-padded, such that all of the spectra in the batch
            are the same length.
        precursors : torch.Tensor of size (n_spectra, 2)
            The measured precursor mass (axis 0) and charge (axis 1) of each
            tandem mass spectrum.

        Returns
        -------
        sequences : list of list of str
            The top sequences for each spectrum.
        scores : torch.Tensor of shape
        (n_spectra, length, n_amino_acids, beam_size)
            The score for each amino acid.
        """
        spectra = spectra.to(self.encoder.device)
        precursors = precursors.to(self.decoder.device)
        scores, tokens = self.beam_search_decode(spectra, precursors)
        sequences = []
        for spec_tokens in tokens:
            sequences.append(
                [self.decoder.detokenize(t) for t in spec_tokens.T]
            )

        return sequences, scores

    def predict_step(self, batch, *args):
        """Sequence a batch of mass spectra.

        Note that this is used within the context of a pytorch-lightning
        Trainer to generate a prediction.

        Parameters
        ----------
        batch : tuple of torch.Tensor
            A batch is expected to contain mass spectra (index 0) and the
            precursor mass and charge (index 1). It may have more indices,
            but these will be ignored.


        Returns
        -------
        sequences : list or str
            The sequence for each spectrum.
        scores : torch.Tensor of shape (n_spectra, length, n_amino_acids)
            The score for each amino acid.
        """
        return self(batch[0], batch[1])

    def beam_search_decode(self, spectra, precursors):
        """Beam search decode the spectra.

        Parameters
        ----------
        spectrum : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra to embed. Axis 0 represents a mass spectrum, axis 1
            contains the peaks in the mass spectrum, and axis 2 is essentially
            a 2-tuple specifying the m/z-intensity pair for each peak. These
            should be zero-padded, such that all of the spectra in the batch
            are the same length.
        precursors : torch.Tensor of size (n_spectra, 2)
            The measured precursor mass (axis 0) and charge (axis 1) of each
            tandem mass spectrum.

        Returns
        -------
        tokens : torch.Tensor of shape
        (n_spectra, max_length, n_amino_acids, beam_size)
            The token sequence for each spectrum.
        scores : torch.Tensor of shape
        (n_spectra, max_length, n_amino_acids, beam_size)
            The score for each amino acid.
        """
        memories, mem_masks = self.encoder(spectra)

        # Sizes:
        batch = spectra.shape[0]  # B
        length = self.max_length + 1  # L
        vocab = self.decoder.vocab_size + 1  # V
        beam = self.beam_size  # S

        # initialize scores and tokens:
        scores = torch.zeros(batch, length, vocab, beam)
        scores = scores.type_as(spectra)
        scores[scores == 0] = torch.nan
        tokens = torch.zeros(batch, length, beam)
        tokens = tokens.type_as(spectra).long()

        # Create the first prediction:
        pred, _ = self.decoder(None, precursors, memories, mem_masks)
        _, idx = torch.topk(pred[:, 0, :], beam, dim=1)
        tokens[:, 0, :] = idx
        scores[:, :1, :, :] = einops.repeat(pred, "B L V -> B L V S", S=beam)

        # Make precursors and memories the right shape for decoding:
        precursors = einops.repeat(precursors, "B L -> (B S) L", S=beam)
        mem_masks = einops.repeat(mem_masks, "B L -> (B S) L", S=beam)
        memories = einops.repeat(memories, "B L V -> (B S) L V", S=beam)

        # The main decoding loop:
        for idx in range(1, self.max_length + 1):
            scores = einops.rearrange(scores, "B L V S -> (B S) L V")
            tokens = einops.rearrange(tokens, "B L S -> (B S) L")
            decoded = (tokens == self.stop_token).any(axis=1)
            if decoded.all():
                scores = einops.rearrange(
                    scores,
                    "(B S) L V -> B L V S",
                    S=beam,
                )
                tokens = einops.rearrange(tokens, "(B S) L -> B L S", S=beam)
                break

            # Get scores:
            scores[~decoded, : idx + 1, :], _ = self.decoder(
                tokens[~decoded, :idx],
                precursors[~decoded, :],
                memories[~decoded, :, :],
                mem_masks[~decoded, :],
            )

            # Reshape to group by spectrum (B for "batch")
            scores = einops.rearrange(scores, "(B S) L V -> B L V S", S=beam)
            tokens = einops.rearrange(tokens, "(B S) L -> B L S", S=beam)
            prev_tokens = einops.repeat(
                tokens[:, :idx, :],
                "B L S -> B L V S",
                V=vocab,
            )

            # Get the previous tokens and scores:
            prev_scores = torch.gather(
                scores[:, :idx, :, :],
                dim=2,
                index=prev_tokens,
            )
            prev_scores = einops.repeat(
                prev_scores[:, :, 0, :],
                "B L S -> B L (V S)",
                V=vocab,
            )

            # Figure out the top K decodings:
            step_scores = torch.zeros(batch, idx + 1, beam * vocab)
            step_scores = step_scores.type_as(scores)
            step_scores[:, :idx, :] = prev_scores
            step_scores[:, idx, :] = einops.rearrange(
                scores[:, idx, :, :],
                "B V S -> B (V S)",
            )

            _, top_idx = torch.topk(step_scores.nanmean(dim=1), beam)
            V_idx, S_idx = np.unravel_index(top_idx, (vocab, beam))

            S_idx = einops.rearrange(S_idx, "B S -> (B S)")
            # V_idx = einops.rearrange(V_idx, "B S -> (B S)")
            B_idx = einops.repeat(torch.arange(batch), "B -> (B S)", S=beam)

            # Record the top K decodings
            tokens[:, :idx, :] = einops.rearrange(
                prev_tokens[B_idx, :, 0, S_idx],
                "(B S) L -> B L S",
                S=beam,
            )
            tokens[:, idx, :] = torch.tensor(V_idx)
            scores[:, : idx + 1, :, :] = einops.rearrange(
                scores[B_idx, : idx + 1, :, S_idx],
                "(B S) L V -> B L V S",
                S=beam,
            )

        return self.softmax(scores), tokens

    def _step(self, spectra, precursors, sequences):
        """The forward learning step.

        Parameters
        ----------
        spectra : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra to embed. Axis 0 represents a mass spectrum, axis 1
            contains the peaks in the mass spectrum, and axis 2 is essentially
            a 2-tuple specifying the m/z-intensity pair for each peak. These
            should be zero-padded, such that all of the spectra in the batch
            are the same length.
        precursors : torch.Tensor of size (n_spectra, 2)
            The measured precursor mass (axis 0) and charge (axis 1) of each
            tandem mass spectrum.
        sequences : list or str of length n_spectra
            The partial peptide sequences to predict.

        Returns
        -------
        scores : torch.Tensor of shape (n_spectra, length, n_amino_acids)
            The raw scores for each amino acid at each position.
        tokens : torch.Tensor of shape (n_spectra, length)
            The best token at each sequence position
        """
        memory, mem_mask = self.encoder(spectra)
        scores, tokens = self.decoder(sequences, precursors, memory, mem_mask)
        return scores, tokens

    def training_step(self, batch, *args):
        """A single training step

        Note that this is used within the context of a pytorch-lightning
        Trainer to generate a prediction.

        Parameters
        ----------
        batch : tuple of torch.Tensor
            A batch is expected to contain mass spectra (index 0), the
            precursor mass and charge (index 1), and the peptide sequence
            (index 2)

        Returns
        -------
        torch.Tensor
            The loss.
        """
        spectra, precursors, sequences = batch
        pred, truth = self._step(spectra, precursors, sequences)
        pred = pred[:, :-1, :].reshape(-1, self.decoder.vocab_size + 1)
        loss = self.celoss(pred, truth.flatten())
        self.log(
            "CELoss",
            {"train": loss.item()},
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )
        return loss

    def validation_step(self, batch, *args):
        """A single validation step

        Note that this is used within the context of a pytorch-lightning
        Trainer to generate a prediction.

        Parameters
        ----------
        batch : tuple of torch.Tensor
            A batch is expected to contain mass spectra (index 0), the
            precursor mass and charge (index 1), and the peptide sequence
            (index 2)

        Returns
        -------
        torch.Tensor
            The loss.
        """
        spectra, precursors, sequences = batch
        pred, truth = self._step(spectra, precursors, sequences)
        pred = pred[:, :-1, :].reshape(-1, self.decoder.vocab_size + 1)
        loss = self.celoss(pred, truth.flatten())
        self.log(
            "CELoss",
            {"valid": loss.item()},
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )
        return loss

    def on_train_epoch_end(self):
        """Log the training loss.

        This is a pytorch-lightning hook.
        """
        metrics = {
            "epoch": self.trainer.current_epoch,
            "train": self.trainer.callback_metrics["CELoss"]["train"].item(),
        }
        self._history.append(metrics)

    def on_validation_epoch_end(self):
        """Log the epoch metrics to self.history.

        This is a pytorch-lightning hook.
        """
        if not self._history:
            return

        valid_loss = self.trainer.callback_metrics["CELoss"]["valid"].item()
        self._history[-1]["valid"] = valid_loss

    def configure_optimizers(self):
        """Initialize the optimizer.

        This is used by pytorch-lightning when preparing the model for
        training.

        Returns
        -------
        torch.optim.Adam
            The intialized Adam optimizer.
        """
        return torch.optim.Adam(self.parameters(), **self.opt_kwargs)



"""Common utility functions"""
import pandas as pd
from tensorboard.backend.event_processing.event_accumulator import (
    EventAccumulator,
)


def read_tensorboard_scalars(path):
    """Read scalars from Tensorboard logs.

    Parameters
    ----------
    path : str
        The path of the scalar log file.

    Returns
    -------
    pandas.DataFrame
        A dataframe containing the scalar values.
    """
    event = EventAccumulator(path)
    event.Reload()
    data = []
    for tag in event.Tags()["scalars"]:
        tag_df = pd.DataFrame(
            event.Scalars(tag), columns=["wall_time", "step", "value"]
        )
        tag_df["tag"] = tag
        data.append(tag_df)

    return pd.concat(data)


def listify(obj):
    """Turn an object into a list, but don't split strings."""
    try:
        assert not isinstance(obj, str)
        iter(obj)
    except (AssertionError, TypeError):
        obj = [obj]

    return list(obj)


# For Parameter Checking ------------------------------------------------------
def check_int(integer, name):
    """Verify that an object is an integer, or coercible to one.

    Parameters
    ----------
    integer : int
        The integer to check.
    name : str
        The name to print in the error message if it fails.
    """
    if isinstance(integer, int):
        return integer

    # Else if it is a float:
    coerced = int(integer)
    if coerced != integer:
        raise ValueError(f"'{name}' must be an integer.")

    return coerced


def check_positive_int(integer, name):
    """Verify that an object is an integer and positive.

    Parameters
    ----------
    integer : int
        The integer to check.
    name : str
        The name to print in the error message if it fails.
    """
    try:
        integer = check_int(integer, name)
        assert integer > 0
    except (ValueError, AssertionError):
        raise ValueError(f"'{name}' must be a positive integer.")

    return integer



"""A PyTorch Dataset class for annotated spectra."""
from typing import Optional, Tuple

import depthcharge
import numpy as np
import spectrum_utils.spectrum as sus
import torch
from torch.utils.data import Dataset


class SpectrumDataset(Dataset):
    """
    Parse and retrieve collections of MS/MS spectra.

    Parameters
    ----------
    spectrum_index : depthcharge.data.SpectrumIndex
        The MS/MS spectra to use as a dataset.
    n_peaks : Optional[int]
        The number of top-n most intense peaks to keep in each spectrum. `None`
        retains all peaks.
    min_mz : float
        The minimum m/z to include. The default is 140 m/z, in order to exclude
        TMT and iTRAQ reporter ions.
    max_mz : float
        The maximum m/z to include.
    min_intensity : float
        Remove peaks whose intensity is below `min_intensity` percentage of the
        base peak intensity.
    remove_precursor_tol : float
        Remove peaks within the given mass tolerance in Dalton around the
        precursor mass.
    random_state : Optional[int]
        The NumPy random state. ``None`` leaves mass spectra in the order they
        were parsed.
    """

    def __init__(
        self,
        spectrum_index: depthcharge.data.SpectrumIndex,
        n_peaks: int = 150,
        min_mz: float = 140.0,
        max_mz: float = 2500.0,
        min_intensity: float = 0.01,
        remove_precursor_tol: float = 2.0,
        random_state: Optional[int] = None,
    ):
        """Initialize a SpectrumDataset"""
        super().__init__()
        self.n_peaks = n_peaks
        self.min_mz = min_mz
        self.max_mz = max_mz
        self.min_intensity = min_intensity
        self.remove_precursor_tol = remove_precursor_tol
        self.rng = np.random.default_rng(random_state)
        self._index = spectrum_index

    def __len__(self) -> int:
        """The number of spectra."""
        return self.n_spectra

    def __getitem__(self, idx) -> Tuple[torch.Tensor, float, int, str]:
        """
        Return the MS/MS spectrum with the given index.

        Parameters
        ----------
        idx : int
            The index of the spectrum to return.

        Returns
        -------
        spectrum : torch.Tensor of shape (n_peaks, 2)
            A tensor of the spectrum with the m/z and intensity peak values.
         ms1: torch.Tensor of shape (5)
             precursor elusion profile
        b_ion: torch.Tensor of shape (peptide_len, vocab_size, n_ion_type, n_neighbors, window_size)
               product ion peptide
        y_ion: torch.Tensor of (peptide_len, vocab_size, n_ion_type, n_neighbors, window_size)
               the product ion peptide
        precursor_mz : float
            The precursor m/z.
        precursor_charge : int
            The precursor charge.
        spectrum_id: str
            The unique spectrum identifier, as determined by its index in the
            original peak file.
        """
        mz_array, int_array, precursor_mz, precursor_charge, ms1, b_ion, y_ion = self.index[idx]
        spectrum = self._process_peaks(
            mz_array, int_array, precursor_mz, precursor_charge
        )
        return spectrum, ms1, b_ion, y_ion, precursor_mz, precursor_charge, str(idx)

    def _process_peaks(
        self,
        mz_array: np.ndarray,
        int_array: np.ndarray,
        precursor_mz: float,
        precursor_charge: int,
    ) -> torch.Tensor:
        """
        Preprocess the spectrum by removing noise peaks and scaling the peak
        intensities.

        Parameters
        ----------
        mz_array : numpy.ndarray of shape (n_peaks,)
            The spectrum peak m/z values.
        int_array : numpy.ndarray of shape (n_peaks,)
            The spectrum peak intensity values.
        precursor_mz : float
            The precursor m/z.
        precursor_charge : int
            The precursor charge.

        Returns
        -------
        torch.Tensor of shape (n_peaks, 2)
            A tensor of the spectrum with the m/z and intensity peak values.
        """
        
        spectrum = sus.MsmsSpectrum(
            "",
            precursor_mz,
            int(precursor_charge),
            mz_array.astype(np.float64),
            int_array.astype(np.float32),
        )
        try:
            spectrum.set_mz_range(self.min_mz, self.max_mz)
            if len(spectrum.mz) == 0:
                raise ValueError
            spectrum.remove_precursor_peak(self.remove_precursor_tol, "Da")
            if len(spectrum.mz) == 0:
                raise ValueError
            spectrum.filter_intensity(self.min_intensity, self.n_peaks)
            if len(spectrum.mz) == 0:
                raise ValueError
            spectrum.scale_intensity("root", 1)
            intensities = spectrum.intensity / np.linalg.norm(
                spectrum.intensity
            )
            return torch.tensor(np.array([spectrum.mz, intensities])).T.float()
        except ValueError:
            # Replace invalid spectra by a dummy spectrum.
            return torch.tensor([[0, 1]]).float()

    @property
    def n_spectra(self) -> int:
        """The total number of spectra."""
        return self.index.n_spectra

    @property
    def index(self) -> depthcharge.data.SpectrumIndex:
        """The underlying SpectrumIndex."""
        return self._index

    @property
    def rng(self):
        """The NumPy random number generator."""
        return self._rng

    @rng.setter
    def rng(self, seed):
        """Set the NumPy random number generator."""
        self._rng = np.random.default_rng(seed)


class AnnotatedSpectrumDataset(SpectrumDataset):
    """
    Parse and retrieve collections of annotated MS/MS spectra.

    Parameters
    ----------
    annotated_spectrum_index : depthcharge.data.SpectrumIndex
        The MS/MS spectra to use as a dataset.
    n_peaks : Optional[int]
        The number of top-n most intense peaks to keep in each spectrum. `None`
        retains all peaks.
    min_mz : float
        The minimum m/z to include. The default is 140 m/z, in order to exclude
        TMT and iTRAQ reporter ions.
    max_mz : float
        The maximum m/z to include.
    min_intensity : float
        Remove peaks whose intensity is below `min_intensity` percentage of the
        base peak intensity.
    remove_precursor_tol : float
        Remove peaks within the given mass tolerance in Dalton around the
        precursor mass.
    random_state : Optional[int]
        The NumPy random state. ``None`` leaves mass spectra in the order they
        were parsed.
    """

    def __init__(
        self,
        annotated_spectrum_index: depthcharge.data.SpectrumIndex,
        n_peaks: int = 150,
        min_mz: float = 140.0,
        max_mz: float = 2500.0,
        min_intensity: float = 0.01,
        remove_precursor_tol: float = 2.0,
        random_state: Optional[int] = None,
    ):
        super().__init__(
            annotated_spectrum_index,
            n_peaks=n_peaks,
            min_mz=min_mz,
            max_mz=max_mz,
            min_intensity=min_intensity,
            remove_precursor_tol=remove_precursor_tol,
            random_state=random_state,
        )

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor,torch.Tensor, torch.Tensor, float, int, str]:
        """
        Return the annotated MS/MS spectrum with the given index.

        Parameters
        ----------
        idx : int
            The index of the spectrum to return.

        Returns
        -------
        spectrum : torch.Tensor of shape (n_peaks, 2)
            A tensor of the spectrum with the m/z and intensity peak values.
        ms1: torch.Tensor of shape (5)
             precursor elusion profile which is intensity values over the retention time range
        b_ion: torch.Tensor of shape (peptide_len, vocab_size, n_ion_type, n_neighbors, window_size)
               product ion peptide
        y_ion: torch.Tensor of (peptide_len, vocab_size, n_ion_type, n_neighbors, window_size)
               the product ion peptide
        precursor_mz : float
            The precursor m/z.
        precursor_charge : int
            The precursor charge.
        annotation : str
            The peptide annotation of the spectrum.
        """
        (
            mz_array,
            int_array,
            precursor_mz,
            precursor_charge,
            ms1,
            ms2_b_ion,
            ms2_y_ion,
            peptide,
        ) = self.index[idx]
    
        spectrum = self._process_peaks(
            mz_array, int_array, precursor_mz, precursor_charge
        )
        return spectrum, ms1, ms2_b_ion, ms2_y_ion, precursor_mz, precursor_charge, peptide



"""The command line entry point for DiffNovo_DIA."""
import datetime
import logging
import re
import sys
import warnings
from typing import Optional, Tuple

warnings.filterwarnings("ignore", category=DeprecationWarning)

import appdirs
import click
import torch
import yaml
from pytorch_lightning.lite import LightningLite

from . import __version__
from . import utils
from .data import ms_io
from .denovo import model_runner

import os

logger = logging.getLogger("diffnovo_dia")


@click.command()
@click.option(
    "--mode",
    required=True,
    default="denovo",
    help="\b\nThe mode in which to run DiffNovo_DIA:\n"
    '- "denovo" will predict peptide sequences for\nunknown MS/MS spectra.\n'
    '- "train" will train a model (from scratch or by\ncontinuing training a '
    "previously trained model).\n"
    '- "eval" will evaluate the performance of a\ntrained model using '
    "previously acquired spectrum\nannotations.",
    type=click.Choice(["denovo", "train", "eval"]),
)
@click.option(
    "--model",
    help="The file name of the model weights (.ckpt file).",
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--peak_path",
    required=True,
    help="The file path with peak files for predicting peptide sequences or "
    "training DiffNovo_DIA.",
)

@click.option(
    "--peak_feature",
    required=True,
    help="The file path with feature files for predicting peptide sequences or "
    "training DiffNovo_DIA.",
)

@click.option(
    "--peak_path_val",
    help="The file path with peak files to be used as validation data during "
    "training.",
)

@click.option(
    "--peak_feature_val",
    help="The file path with peak feature files to be used as validation data during "
    "training.",
)


@click.option(
    "--config",
    help="The file name of the configuration file with custom options. If not "
    "specified, a default configuration will be used.",
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--output",
    help="The base output file name to store logging (extension: .log) and "
    "(optionally) prediction results (extension: .csv).",
    type=click.Path(dir_okay=False),
)
def main(
    mode: str,
    model: Optional[str],
    peak_path: str,
    peak_feature: str,
    peak_path_val: Optional[str],
    peak_feature_val: Optional[str],
    config: Optional[str],
    output: Optional[str],
):
    """
    \b
    DiffNovo-DIA:Transformer-Diffusion de novo peptide sequencing for data-independent acquisition mass spectrometry.
    ================================================================================

    Ebrahimi, S. DE NOVO PEPTIDE SEQUENCING FOR DATA-INDEPENDENT ACQUISITION (DIA) USING DEEP LEARNING
    https://github.com/Biocomputing-Research-Group/DiffNovo-DIA
    """
    if output is None:
        output = os.path.join(
            os.getcwd(),
            f"transdia_{datetime.datetime.now().strftime('%Y%m%d%H%M%S')}",
        )
    else:
        output = os.path.splitext(os.path.abspath(output))[0]

    # Configure logging.
    logging.captureWarnings(True)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    log_formatter = logging.Formatter(
        "{asctime} {levelname} [{name}/{processName}] {module}.{funcName} : "
        "{message}",
        style="{",
    )
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setLevel(logging.DEBUG)
    console_handler.setFormatter(log_formatter)
    root.addHandler(console_handler)
    file_handler = logging.FileHandler(f"{output}.log")
    file_handler.setFormatter(log_formatter)
    root.addHandler(file_handler)
    # Disable dependency non-critical log messages.
    logging.getLogger("depthcharge").setLevel(logging.INFO)
    logging.getLogger("github").setLevel(logging.WARNING)
    logging.getLogger("h5py").setLevel(logging.WARNING)
    logging.getLogger("numba").setLevel(logging.WARNING)
    logging.getLogger("pytorch_lightning").setLevel(logging.WARNING)
    logging.getLogger("torch").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)

    # Read parameters from the config file.
    if config is None:
        config = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "config.yaml"
        )
    config_fn = config
    with open(config) as f_in:
        config = yaml.safe_load(f_in)
    # Ensure that the config values have the correct type.
    config_types = dict(
        random_seed=int,
        n_peaks=int,
        min_mz=float,
        max_mz=float,
        min_intensity=float,
        remove_precursor_tol=float,
        max_charge=int,
        precursor_mass_tol=float,
        isotope_error_range=lambda min_max: (int(min_max[0]), int(min_max[1])),
        dim_model=int,
        vocab_size = int,
        win_size = int,
        num_ions = int,
        num_neighbors = int,
        n_head=int,
        dim_feedforward=int,
        n_layers=int,
        dropout=float,
        dim_intensity=int,
        max_length=int,
        n_log=int,
        warmup_iters=int,
        max_iters=int,
        learning_rate=float,
        weight_decay=float,
        train_batch_size=int,
        predict_batch_size=int,
        n_beams=int,
        max_epochs=int,
        num_sanity_val_steps=int,
        train_from_scratch=bool,
        save_model=bool,
        model_save_folder_path=str,
        save_weights_only=bool,
        every_n_train_steps=int,
    )
    for k, t in config_types.items():
        try:
            if config[k] is not None:
                config[k] = t(config[k])
        except (TypeError, ValueError) as e:
            logger.error("Incorrect type for configuration value %s: %s", k, e)
            raise TypeError(f"Incorrect type for configuration value {k}: {e}")
    config["residues"] = {
        str(aa): float(mass) for aa, mass in config["residues"].items()
    }
    # Add extra configuration options and scale by the number of GPUs.
    n_gpus = torch.cuda.device_count()
    config["n_workers"] = utils.n_workers()
    if n_gpus > 1:
        config["train_batch_size"] = config["train_batch_size"] // n_gpus

    LightningLite.seed_everything(seed=config["random_seed"], workers=True)
    if model is None and mode != "train":

            raise PermissionError(
                "GitHub API rate limit exceeded while trying to download the "
                "model weights"
            ) from None

    # Log the active configuration.
    logger.info("DiffNovo_DIA version %s", str(__version__))
    logger.debug("mode = %s", mode)
    logger.debug("model = %s", model)
    logger.debug("peak_path = %s", peak_path)
    logger.debug("peak_feature= %s", peak_feature)
    logger.debug("peak_path_val = %s", peak_path_val)
    logger.debug("peak_feature_val = %s", peak_feature_val)
    logger.debug("config = %s", config_fn)
    logger.debug("output = %s", output)
    for key, value in config.items():
        logger.debug("%s = %s", str(key), str(value))

    # Run DiffNovo_DIA in the specified mode.
    if mode == "denovo":
        logger.info("Predict peptide sequences with DiffNovo_DIA.")
        writer = ms_io.MztabWriter(f"{output}.csv")

        model_runner.predict(peak_path, peak_feature, model, config, writer)
        writer.save()
    elif mode == "eval":
        writer = ms_io.MztabWriter(f"{output}.csv")
        logger.info("Evaluate a trained DiffNovo_DIA model.")
        model_runner.evaluate(peak_path, peak_feature, model, config, writer)
        writer.save()
    elif mode == "train":
        logger.info("Train the DiffNovo_DIA model.")
        model_runner.train(peak_path, peak_feature, peak_path_val, peak_feature_val, model, config)

if __name__ == "__main__":
    main()



"""Useful mixins for model classes"""
import pandas as pd


class ModelMixin:
    """Add some useful methods for models."""

    def init(self):
        """Initialize the ModelMixin"""
        self._history = []

    @property
    def history(self):
        """The training history of a model."""
        return pd.DataFrame(self._history)

    @property
    def n_parameters(self):
        """The number of learnable parameters."""
        return sum(p.numel() for p in self.parameters() if p.requires_grad)



"""Methods to evaluate peptide-spectrum predictions."""
import re
from typing import Dict, Iterable, List, Tuple

import numpy as np
from spectrum_utils.utils import mass_diff


def aa_match_prefix(
    peptide1: List[str],
    peptide2: List[str],
    aa_dict: Dict[str, float],
    cum_mass_threshold: float = 0.5,
    ind_mass_threshold: float = 0.1,
) -> Tuple[np.ndarray, bool]:
    """
    Find the matching prefix amino acids between two peptide sequences.

    This is a similar evaluation criterion as used by DeepNovo.

    Parameters
    ----------
    peptide1 : List[str]
        The first tokenized peptide sequence to be compared.
    peptide2 : List[str]
        The second tokenized peptide sequence to be compared.
    aa_dict : Dict[str, float]
        Mapping of amino acid tokens to their mass values.
    cum_mass_threshold : float
        Mass threshold in Dalton to accept cumulative mass-matching amino acid
        sequences.
    ind_mass_threshold : float
        Mass threshold in Dalton to accept individual mass-matching amino acids.

    Returns
    -------
    aa_matches : np.ndarray of length max(len(peptide1), len(peptide2))
        Boolean flag indicating whether each paired-up amino acid matches across
        both peptide sequences.
    pep_match : bool
        Boolean flag to indicate whether the two peptide sequences fully match.
    """
    aa_matches = np.zeros(max(len(peptide1), len(peptide2)), np.bool_)
    # Find longest mass-matching prefix.
    i1, i2, cum_mass1, cum_mass2 = 0, 0, 0.0, 0.0
    while i1 < len(peptide1) and i2 < len(peptide2):
        aa_mass1 = aa_dict.get(peptide1[i1], 0)
        aa_mass2 = aa_dict.get(peptide2[i2], 0)
        if (
            abs(mass_diff(cum_mass1 + aa_mass1, cum_mass2 + aa_mass2, True))
            < cum_mass_threshold
        ):
            aa_matches[max(i1, i2)] = (
                abs(mass_diff(aa_mass1, aa_mass2, True)) < ind_mass_threshold
            )
            i1, i2 = i1 + 1, i2 + 1
            cum_mass1, cum_mass2 = cum_mass1 + aa_mass1, cum_mass2 + aa_mass2
        elif cum_mass2 + aa_mass2 > cum_mass1 + aa_mass1:
            i1, cum_mass1 = i1 + 1, cum_mass1 + aa_mass1
        else:
            i2, cum_mass2 = i2 + 1, cum_mass2 + aa_mass2
    return aa_matches, aa_matches.all()


def aa_match_prefix_suffix(
    peptide1: List[str],
    peptide2: List[str],
    aa_dict: Dict[str, float],
    cum_mass_threshold: float = 0.5,
    ind_mass_threshold: float = 0.1,
) -> Tuple[np.ndarray, bool]:
    """
    Find the matching prefix and suffix amino acids between two peptide
    sequences.

    Parameters
    ----------
    peptide1 : List[str]
        The first tokenized peptide sequence to be compared.
    peptide2 : List[str]
        The second tokenized peptide sequence to be compared.
    aa_dict : Dict[str, float]
        Mapping of amino acid tokens to their mass values.
    cum_mass_threshold : float
        Mass threshold in Dalton to accept cumulative mass-matching amino acid
        sequences.
    ind_mass_threshold : float
        Mass threshold in Dalton to accept individual mass-matching amino acids.

    Returns
    -------
    aa_matches : np.ndarray of length max(len(peptide1), len(peptide2))
        Boolean flag indicating whether each paired-up amino acid matches across
        both peptide sequences.
    pep_match : bool
        Boolean flag to indicate whether the two peptide sequences fully match.
    """
    # Find longest mass-matching prefix.
    aa_matches, pep_match = aa_match_prefix(
        peptide1, peptide2, aa_dict, cum_mass_threshold, ind_mass_threshold
    )
    # No need to evaluate the suffixes if the sequences already fully match.
    if pep_match:
        return aa_matches, pep_match
    # Find longest mass-matching suffix.
    i1, i2 = len(peptide1) - 1, len(peptide2) - 1
    i_stop = np.argwhere(~aa_matches)[0]
    cum_mass1, cum_mass2 = 0.0, 0.0
    while i1 >= i_stop and i2 >= i_stop:
        aa_mass1 = aa_dict.get(peptide1[i1], 0)
        aa_mass2 = aa_dict.get(peptide2[i2], 0)
        if (
            abs(mass_diff(cum_mass1 + aa_mass1, cum_mass2 + aa_mass2, True))
            < cum_mass_threshold
        ):
            aa_matches[max(i1, i2)] = (
                abs(mass_diff(aa_mass1, aa_mass2, True)) < ind_mass_threshold
            )
            i1, i2 = i1 - 1, i2 - 1
            cum_mass1, cum_mass2 = cum_mass1 + aa_mass1, cum_mass2 + aa_mass2
        elif cum_mass2 + aa_mass2 > cum_mass1 + aa_mass1:
            i1, cum_mass1 = i1 - 1, cum_mass1 + aa_mass1
        else:
            i2, cum_mass2 = i2 - 1, cum_mass2 + aa_mass2
    return aa_matches, aa_matches.all()


def aa_match(
    peptide1: List[str],
    peptide2: List[str],
    aa_dict: Dict[str, float],
    cum_mass_threshold: float = 0.5,
    ind_mass_threshold: float = 0.1,
    mode: str = "best",
) -> Tuple[np.ndarray, bool]:
    """
    Find the matching amino acids between two peptide sequences.

    Parameters
    ----------
    peptide1 : List[str]
        The first tokenized peptide sequence to be compared.
    peptide2 : List[str]
        The second tokenized peptide sequence to be compared.
    aa_dict : Dict[str, float]
        Mapping of amino acid tokens to their mass values.
    cum_mass_threshold : float
        Mass threshold in Dalton to accept cumulative mass-matching amino acid
        sequences.
    ind_mass_threshold : float
        Mass threshold in Dalton to accept individual mass-matching amino acids.
    mode : {"best", "forward", "backward"}
        The direction in which to find matching amino acids.

    Returns
    -------
    aa_matches : np.ndarray of length max(len(peptide1), len(peptide2))
        Boolean flag indicating whether each paired-up amino acid matches across
        both peptide sequences.
    pep_match : bool
        Boolean flag to indicate whether the two peptide sequences fully match.
    """
    if mode == "best":
        return aa_match_prefix_suffix(
            peptide1, peptide2, aa_dict, cum_mass_threshold, ind_mass_threshold
        )
    elif mode == "forward":
        return aa_match_prefix(
            peptide1, peptide2, aa_dict, cum_mass_threshold, ind_mass_threshold
        )
    elif mode == "backward":
        aa_matches, pep_match = aa_match_prefix(
            list(reversed(peptide1)),
            list(reversed(peptide2)),
            aa_dict,
            cum_mass_threshold,
            ind_mass_threshold,
        )
        return aa_matches[::-1], pep_match
    else:
        raise ValueError("Unknown evaluation mode")


def aa_match_batch(
    peptides1: Iterable,
    peptides2: Iterable,
    aa_dict: Dict[str, float],
    cum_mass_threshold: float = 0.5,
    ind_mass_threshold: float = 0.1,
    mode: str = "best",
) -> Tuple[List[Tuple[np.ndarray, bool]], int, int]:
    """
    Find the matching amino acids between multiple pairs of peptide sequences.

    Parameters
    ----------
    peptides1 : Iterable
        The first list of peptide sequences to be compared.
    peptides2 : Iterable
        The second list of peptide sequences to be compared.
    aa_dict : Dict[str, float]
        Mapping of amino acid tokens to their mass values.
    cum_mass_threshold : float
        Mass threshold in Dalton to accept cumulative mass-matching amino acid
        sequences.
    ind_mass_threshold : float
        Mass threshold in Dalton to accept individual mass-matching amino acids.
    mode : {"best", "forward", "backward"}
        The direction in which to find matching amino acids.

    Returns
    -------
    aa_matches_batch : List[Tuple[np.ndarray, bool]]
        For each pair of peptide sequences: (i) boolean flags indicating whether
        each paired-up amino acid matches across both peptide sequences, (ii)
        boolean flag to indicate whether the two peptide sequences fully match.
    n_aa1: int
        Total number of amino acids in the first list of peptide sequences.
    n_aa2: int
        Total number of amino acids in the second list of peptide sequences.
    """
    aa_matches_batch, n_aa1, n_aa2 = [], 0, 0
    for peptide1, peptide2 in zip(peptides1, peptides2):
        # Split peptides into individual AAs if necessary.
        if isinstance(peptide1, str):
            peptide1 = re.split(r"(?<=.)(?=[A-Z])", peptide1)
        if isinstance(peptide2, str):
            peptide2 = re.split(r"(?<=.)(?=[A-Z])", peptide2)
        n_aa1, n_aa2 = n_aa1 + len(peptide1), n_aa2 + len(peptide2)
        aa_matches_batch.append(
            aa_match(
                peptide1,
                peptide2,
                aa_dict,
                cum_mass_threshold,
                ind_mass_threshold,
                mode,
            )
        )
    return aa_matches_batch, n_aa1, n_aa2


def aa_match_metrics(
    aa_matches_batch: List[Tuple[np.ndarray, bool]],
    n_aa_true: int,
    n_aa_pred: int,
) -> Tuple[float, float, float]:
    """
    Calculate amino acid and peptide-level evaluation metrics.

    Parameters
    ----------
    aa_matches_batch : List[Tuple[np.ndarray, bool]]
        For each pair of peptide sequences: (i) boolean flags indicating whether
        each paired-up amino acid matches across both peptide sequences, (ii)
        boolean flag to indicate whether the two peptide sequences fully match.
    n_aa_true: int
        Total number of amino acids in the true peptide sequences.
    n_aa_pred: int
        Total number of amino acids in the predicted peptide sequences.

    Returns
    -------
    aa_precision: float
        The number of correct AA predictions divided by the number of predicted
        AAs.
    aa_recall: float
        The number of correct AA predictions divided by the number of true AAs.
    pep_recall: float
        The number of correct peptide predictions divided by the number of
        peptides.
    """
    n_aa_correct = sum(
        [aa_matches[0].sum() for aa_matches in aa_matches_batch]
    )
    aa_precision = n_aa_correct / (n_aa_pred + 1e-8)
    aa_recall = n_aa_correct / (n_aa_true + 1e-8)
    pep_recall = sum([aa_matches[1] for aa_matches in aa_matches_batch]) / (
        len(aa_matches_batch) + 1e-8
    )
    return aa_precision, aa_recall, pep_recall


def aa_precision_recall(
    aa_scores_correct: List[float],
    aa_scores_all: List[float],
    n_aa_total: int,
    threshold: float,
) -> Tuple[float, float]:
    """
    Calculate amino acid level precision and recall at a given score threshold.

    Parameters
    ----------
    aa_scores_correct : List[float]
        Amino acids scores for the correct amino acids predictions.
    aa_scores_all : List[float]
        Amino acid scores for all amino acids predictions.
    n_aa_total : int
        The total number of amino acids in the predicted peptide sequences.
    threshold : float
        The amino acid score threshold.

    Returns
    -------
    aa_precision: float
        The number of correct amino acid predictions divided by the number of
        predicted amino acids.
    aa_recall: float
        The number of correct amino acid predictions divided by the total number
        of amino acids.
    """
    n_aa_correct = sum([score > threshold for score in aa_scores_correct])
    n_aa_predicted = sum([score > threshold for score in aa_scores_all])
    return n_aa_correct / n_aa_predicted, n_aa_correct / n_aa_total



MAX_LEN=30
import os
import pickle
import re
import numpy as np
from collections import defaultdict
import argparse

class data_preprocess(object):
  """
  This is a helper class designed for multi-process get_spectrum
  """

  def dict_to_pkl(self, fname, dict):

      with open(fname, mode='wb') as handle:
          pickle.dump(dict, handle, pickle.HIGHEST_PROTOCOL)

  def __init__(self, input_spectrum_file, input_feature_file):
    self.MZ_MAX = 2500
    self.SPECTRUM_RESOLUTION = 50
    self.MZ_SIZE = MZ_SIZE = int(self.MZ_MAX * self.SPECTRUM_RESOLUTION)
    self.neighbor_size = 5

    self.dia_window = 20.0 # the window size of MS2 scan in Dalton

    self.input_spectrum_file = input_spectrum_file
    self.input_feature_file = input_feature_file
    self.output_file = ''

    # split data into batches
    self.feature_index_list =[]


    ### store file location of each feature for random access
    self.feature_location_list = []

    # store the file location of all spectra for random access
    self.spectrum_location_dict =  {}
    self.spectrum_rtinseconds_dict = {}

    # record the status of spectra that have been read
    self.feature_count = {"total": 0,
                          "read": 0,
                          "skipped": 0,
                          "skipped_mass": 0}
    self.spectrum_count = 0
    self.precursor_info = defaultdict(list)
    self.col_feature_id = 0
    self.col_precursor_mz = 1
    self.col_precursor_charge = 2
    self.col_rt_mean = 3
    self.col_raw_sequence = 4
    self.col_scan_list = 5
    self.col_ms1_list = 6
    self.col_feature_area = 7
    self.col_num = 8
    # predicted file column format
    self.pcol_feature_id = 0
    self.pcol_feature_area = 1
    self.pcol_sequence = 2
    self.pcol_score = 3
    self.pcol_position_score = 4
    self.pcol_precursor_mz = 5
    self.pcol_precursor_charge = 6
    self.pcol_protein_id = 7
    self.pcol_scan_list_middle = 8
    self.pcol_scan_list_original = 9
    self.pcol_score_max = 10
    self.mass_H = 1.0078
    self.mass_H2O = 18.0106
    self.mass_NH3 = 17.0265
    self.mass_N_terminus = 1.0078
    self.mass_C_terminus = 17.0027
    self.mass_CO = 27.9949

  def get_location(self):

    print("".join(["="] * 80)) # section-separating line
    print("Feature Extraction: get_location()")

    ### store file location of each spectrum for random access {scan:location}
    ### since mgf file can be rather big, cache the locations for each spectrum mgf file.
    spectrum_location_file = self.input_spectrum_file + '.locations.pkl'
    if os.path.exists(spectrum_location_file):
      print("Feature Extraction: read cached spectrum locations")
      with open(spectrum_location_file, 'rb') as fr:
        data = pickle.load(fr)
        self.spectrum_location_dict, self.spectrum_rtinseconds_dict, self.spectrum_count = data
    else:
      print("Feature Extraction: build spectrum location from scratch")
      spectrum_location_dict = {}
      spectrum_rtinseconds_dict = {}
      line = True
      while line:
        current_location = self.input_spectrum_handle.tell()
        line = self.input_spectrum_handle.readline()
        if "BEGIN IONS" in line:
          spectrum_location = current_location
        elif "SCANS=" in line:
          scan = re.split('=|\r\n', line)[1]
          spectrum_location_dict[scan] = spectrum_location
        elif "RTINSECONDS=" in line:
          rtinseconds = float(re.split('=|\r\n', line)[1])
          spectrum_rtinseconds_dict[scan] = rtinseconds
      self.spectrum_location_dict = spectrum_location_dict
      self.spectrum_rtinseconds_dict = spectrum_rtinseconds_dict
      self.spectrum_count = len(spectrum_location_dict)
      with open(spectrum_location_file, 'wb') as fw:
        pickle.dump((self.spectrum_location_dict, self.spectrum_rtinseconds_dict, self.spectrum_count), fw)

    ### store file location of each feature for random access
    # skip header line
    _ = self.input_feature_handle.readline()
    line = True
    while line:
      feature_location = self.input_feature_handle.tell()
      self.feature_location_list.append(feature_location)
      line = self.input_feature_handle.readline()
    self.feature_location_list = self.feature_location_list[:-1]
    self.feature_location_list = self.feature_location_list
    self.feature_count["total"] = len(self.feature_location_list)
    self.feature_index_list = range(self.feature_count["total"])

    print("spectrum_count = {0:d}".format(self.spectrum_count))
    print("feature_count[total] = {0:d}".format(self.feature_count["total"]))

  def open_input(self):

    print("".join(["="] * 80)) # section-separating line
    print("Feature Extraction: open_input()")

    self.input_spectrum_handle = open(self.input_spectrum_file, 'r')
    self.input_feature_handle = open(self.input_feature_file, 'r')

  def _print_prediction_header(self):
    print("".join(["="] * 80)) # section-separating line
    print("WorkerIO: _print_prediction_header()")

    header_list = ["feature_id",
                   "feature_area",
                   "predicted_sequence",
                   "predicted_score",
                   "predicted_position_score",
                   "precursor_mz",
                   "precursor_charge",
                   "protein_access_id",
                   "scan_list_middle",
                   "scan_list_original",
                   "predicted_score_max"]
    header_row = "\t".join(header_list)
    print(header_row, file=self.output_handle, end="\n")
  def open_output(self):

    print("".join(["="] * 80)) # section-separating line
    print("Feature Extraction: open_output()")

    self.output_handle = open(self.output_file, 'w')
    self._print_prediction_header()


  def get_spectrum(self, feature_index_batch, input_feature_file_handle, input_spectrum_file_handle):
    spectrum_list = []
    for feature_index in feature_index_batch:
      # parse a feature
      feature_location = self.feature_location_list[feature_index]
      feature_id, feature_area, precursor_mz, precursor_charge, rt_mean, raw_sequence, scan_list, ms1_list = self._parse_feature(feature_location, input_feature_file_handle)
      # skip if precursor_mass > MZ_MAX
      precursor_mass = precursor_mz * precursor_charge - self.mass_H * precursor_charge
      if precursor_mass > self.MZ_MAX:
        continue

      # parse and process spectrum
      (neighbor_right_count,
       neighbor_size_half,
       scan_list_middle,
        mz_list,
       intensity_list,
       ms1_profile) = self._parse_spectrum(precursor_mz, precursor_mass, rt_mean, scan_list, ms1_list, input_spectrum_file_handle)
      # update dataset
      spectrum = {
                  "raw_sequence": raw_sequence,
                  "precursor_mass": precursor_mass,
                  "precursor_mz": precursor_mz,
                  "precursor_charge": precursor_charge,
                  "ms1_profile": ms1_profile,
                  "mz_list":mz_list,
                  "intensity_list": intensity_list,
                  "scan_list_middle": scan_list_middle,
                  "neighbor_right_count":neighbor_right_count,
                  "neighbor_size_half":neighbor_size_half}
      spectrum_list.append(spectrum)

    return spectrum_list

  def _parse_feature(self, feature_location, input_file_handle):
    input_file_handle.seek(feature_location)
    line = input_file_handle.readline()
    line = re.split(',|\r|\n', line)
    feature_id = line[self.col_feature_id]
    feature_area = 0#float(line[deepnovo_config.col_feature_area])
    precursor_mz = float(line[self.col_precursor_mz])
    precursor_charge = float(line[self.col_precursor_charge])
    rt_mean = float(line[self.col_rt_mean])
    raw_sequence = line[self.col_raw_sequence]
    scan_list = re.split(';', line[self.col_scan_list])
    ms1_list = re.split(';', line[self.col_ms1_list])
    assert len(scan_list) == len(ms1_list), "Error: scan_list and ms1_list not matched."

    return feature_id, feature_area, precursor_mz, precursor_charge, rt_mean, raw_sequence, scan_list, ms1_list

  def _parse_spectrum_ion(self, input_file_handle):
    mz_list = []
    intensity_list = []
    line = input_file_handle.readline()
    while not "END IONS" in line:
      mz, intensity = re.split(' |\n', line)[:2]
      mz_float = float(mz)
      intensity_float = float(intensity)
      # skip an ion if its mass > MZ_MAX
      if mz_float > self.MZ_MAX:
        line = input_file_handle.readline()
        continue
      mz_list.append(mz_float)
      intensity_list.append(intensity_float)
      line = input_file_handle.readline()

    return mz_list, intensity_list

  def _parse_spectrum(self, precursor_mz, precursor_mass, rt_mean, scan_list, ms1_list, input_file_handle):
    mz_five_list = []
    int_five_list = []
    neighbor_count = len(scan_list)
    best_scan_index = None
    best_distance = float('inf')
    for scan_index, scan in enumerate(scan_list):
      try:
        distance = abs(self.spectrum_rtinseconds_dict[scan] - rt_mean)
      except:
        distance = abs(self.spectrum_rtinseconds_dict[scan+ '\n'] - rt_mean)
      if distance < best_distance:
        best_distance = distance
        best_scan_index = scan_index
    neighbor_center = best_scan_index
    neighbor_left_count = neighbor_center
    neighbor_right_count = neighbor_count - neighbor_left_count - 1
    neighbor_size_half = self.neighbor_size // 2
    neighbor_left_count = min(neighbor_left_count, neighbor_size_half)
    neighbor_right_count = min(neighbor_right_count, neighbor_size_half)
    ### parse and add neighbor spectra
    scan_list_middle = []
    ms1_intensity_list_middle = []
    for index in range(neighbor_center - neighbor_left_count, neighbor_center + neighbor_right_count + 1):
      scan = scan_list[index]
      scan_list_middle.append(scan)
      ms1_entry = ms1_list[index]
      ms1_intensity = float(re.split(':', ms1_entry)[1])
      ms1_intensity_list_middle.append(ms1_intensity)
    ms1_intensity_max = max(ms1_intensity_list_middle)
    assert ms1_intensity_max > 0.0, "Error: Zero ms1_intensity_max"
    ms1_intensity_list_middle = [x/ms1_intensity_max for x in ms1_intensity_list_middle]
    for scan, ms1_intensity in zip(scan_list_middle, ms1_intensity_list_middle):
      try:
          spectrum_location = self.spectrum_location_dict[scan]
      except:
          spectrum_location = self.spectrum_location_dict[scan+'\n']
      input_file_handle.seek(spectrum_location)
      # parse header lines
      line = input_file_handle.readline()
      assert "BEGIN IONS" in line, "Error: wrong input BEGIN IONS"
      line = input_file_handle.readline()
      assert "TITLE=" in line, "Error: wrong input TITLE="
      line = input_file_handle.readline()
      assert "PEPMASS=" in line, "Error: wrong input PEPMASS="
      line = input_file_handle.readline()
      assert "CHARGE=" in line, "Error: wrong input CHARGE="
      line = input_file_handle.readline()
      assert "SCANS=" in line, "Error: wrong input SCANS="
      line = input_file_handle.readline()
      assert "RTINSECONDS=" in line, "Error: wrong input RTINSECONDS="
      # parse fragment ions
      mz_list, intensity_list = self._parse_spectrum_ion(input_file_handle)
      mz_five_list.append(mz_list)
      int_five_list.append(intensity_list)

    # ms1_profile
    for x in range(neighbor_size_half - neighbor_left_count):
      ms1_intensity_list_middle = [0.0] + ms1_intensity_list_middle
    for x in range(neighbor_size_half - neighbor_right_count):
      ms1_intensity_list_middle = ms1_intensity_list_middle + [0.0]
    assert len(ms1_intensity_list_middle) == self.neighbor_size, "Error: ms1 profile"
    ms1_profile = np.array(ms1_intensity_list_middle)

    return neighbor_right_count, neighbor_size_half, scan_list_middle, mz_five_list, int_five_list, ms1_profile

  def read_precursor_info(self, feature_index, feature_fr, spectrum_fr):

      spectrum_list = self.get_spectrum([feature_index], feature_fr, spectrum_fr)
      if not spectrum_list:
          return None

      spectrum = spectrum_list[0]
      raw_sequence = spectrum["raw_sequence"]
      spec_info = {
          "precursor_mz": spectrum["precursor_mz"],
          "precursor_charge": spectrum["precursor_charge"],
          "scan_list_middle": spectrum["scan_list_middle"],
          "ms1": spectrum["ms1_profile"],
          "mz_list": spectrum["mz_list"],
          "int_list": spectrum["intensity_list"],
          "neighbor_right_count": spectrum["neighbor_right_count"],
          "neighbor_size_half": spectrum["neighbor_size_half"]}
      ptms = ['C(+57.02)']
      ptm_casanovo_format = 'C+57.021'
      for ptm in ptms:

          if ptm in raw_sequence:
              raw_sequence = raw_sequence.replace(ptm, ptm_casanovo_format)
      self.precursor_info[raw_sequence].append(spec_info)
      return
  def get_precursor_info(self):
      with open(self.input_feature_file, 'r') as feature_fr:
          with open(self.input_spectrum_file, 'r') as spectrum_fr:
              for feature_index in self.feature_index_list:
                  self.read_precursor_info(feature_index, feature_fr, spectrum_fr)




if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process spectrum and feature files to generate a feature file in pickle format for Transformer_DIA or DiffNovo_DIA.")
    parser.add_argument("--feature_file", required=True, help="Path to the feature file (.csv)")
    parser.add_argument("--spectrum_file", required=True, help="Path to the spectrum file (.mgf)")
    parser.add_argument("--output_file", required=True, help="Path to save the output pickle file (.pkl)")

    args = parser.parse_args()

    # Validate input files
    if not os.path.isfile(args.feature_file) or not args.feature_file.endswith(".csv"):
        raise ValueError("Feature file must exist and have a .csv extension.")
    if not os.path.isfile(args.spectrum_file) or not args.spectrum_file.endswith(".mgf"):
        raise ValueError("Spectrum file must exist and have a .mgf extension.")

    print(
        "This script processes spectrum and feature files to generate a feature file in pickle format, "
        "required for the Transformer_DIA or DiffNovo_DIA model.\n"
        "Generated features:\n"
        "- Keys: Peptide sequences\n"
        "- Values: [precursor_mz, precursor_charge, scan_list_middle, ms1, mz_list, int_list, neighbor_right_count, neighbor_size_half]"
    )

    # Run preprocessing
    feature_extraction = data_preprocess(args.spectrum_file, args.feature_file)
    feature_extraction.open_input()
    feature_extraction.get_location()
    feature_extraction.get_precursor_info()
    feature_extraction.dict_to_pkl(args.output_file, feature_extraction.precursor_info)




import numpy as np

# Special vocabulary symbols - we always put them at the start.
_PAD = "_PAD"
_GO = "_GO"
_EOS = "_EOS"
_START_VOCAB = [_PAD, _GO, _EOS]

PAD_ID = 0
GO_ID = 1
EOS_ID = 2
vocab_reverse = ['A',
                 'R',
                 'N',
                 'N(Deamidation)',
                 'D',
                 # ~ 'C',
                 'C(Carbamidomethylation)',
                 'E',
                 'Q',
                 'Q(Deamidation)',
                 'G',
                 'H',
                 'I',
                 'L',
                 'K',
                 'M',
                 'M(Oxidation)',
                 'F',
                 'P',
                 'S',
                 'T',
                 'W',
                 'Y',
                 'V',
                 ]

vocab_reverse = _START_VOCAB + vocab_reverse

vocab = dict([(x, y) for (y, x) in enumerate(vocab_reverse)])

vocab_size = len(vocab_reverse)
mass_H = 1.0078
mass_H2O = 18.0106
mass_NH3 = 17.0265
mass_N_terminus = 1.0078
mass_C_terminus = 17.0027
mass_CO = 27.9949

mass_AA = {'_PAD': 0.0,
           '_GO': mass_N_terminus - mass_H,
           '_EOS': mass_C_terminus + mass_H,
           'A': 71.03711,  # 0
           'R': 156.10111,  # 1
           'N': 114.04293,  # 2
           'N(Deamidation)': 115.02695,
           'D': 115.02694,  # 3
           # ~ 'C(Carbamidomethylation)': 103.00919, # 4
           'C(Carbamidomethylation)': 160.03065,  # C(+57.02)
           # ~ 'C(Carbamidomethylation)': 161.01919, # C(+58.01) # orbi
           'E': 129.04259,  # 5
           'Q': 128.05858,  # 6
           'Q(Deamidation)': 129.0426,
           'G': 57.02146,  # 7
           'H': 137.05891,  # 8
           'I': 113.08406,  # 9
           'L': 113.08406,  # 10
           'K': 128.09496,  # 11
           'M': 131.04049,  # 12
           'M(Oxidation)': 147.0354,
           'F': 147.06841,  # 13
           'P': 97.05276,  # 14
           'S': 87.03203,  # 15
           'T': 101.04768,  # 16
           'W': 186.07931,  # 17
           'Y': 163.06333,  # 18
           'V': 99.06841,  # 19
           }

mass_ID = [mass_AA[vocab_reverse[x]] for x in range(vocab_size)]
mass_ID_np = np.array(mass_ID, dtype=np.float32)

mass_AA_min = mass_AA["G"]  # 57.02146
num_ion = 8


class WorkerDIA(object):
    def __init__(self):
        self.MZ_MAX = 3000
        self.SPECTRUM_RESOLUTION = 50  # bins for 1.0 Da = precision 0.02 Da
        self.MZ_SIZE = int(self.MZ_MAX * self.SPECTRUM_RESOLUTION)
        self.neighbor_size = 5
        self.mass_H = 1.0078
        self.dia_window = 10
        self._buckets = [12, 22, 32]
        self.vocab = vocab
        self.GO_ID = GO_ID
        self.EOS_ID = EOS_ID
        self.PAD_ID = PAD_ID
        self.vocab_size = vocab_size
        self.WINDOW_SIZE = self.dia_window
        self.SPECTRUM_RESOLUTION = self.SPECTRUM_RESOLUTION
        self.mass_ID_np = mass_ID_np
        self.mass_H2O = mass_H2O
        self.mass_NH3 = mass_NH3
        self.num_ion = num_ion
        self.mass_C_terminus = mass_C_terminus
        self.mass_N_terminus = mass_N_terminus
        self.mass_ID = mass_ID

    def process_spectrum(self, spectrum_mz_list, spectrum_intensity_list, peptide_mass):

        # neutral mass, location, assuming ion charge z=1
        charge = 1.0
        spectrum_mz = np.array(spectrum_mz_list, dtype=np.float32)
        neutral_mass = spectrum_mz - charge * self.mass_H
        neutral_mass_location = np.rint(neutral_mass * self.SPECTRUM_RESOLUTION).astype(
            np.int32)

        neutral_mass_location_view = neutral_mass_location

        # intensity
        spectrum_intensity = np.array(spectrum_intensity_list, dtype=np.float32)
        spectrum_intensity_max = np.max(spectrum_intensity)
        norm_intensity = spectrum_intensity
        norm_intensity_view = norm_intensity

        # fill spectrum holders
        spectrum_holder = np.zeros(shape=(1, self.MZ_SIZE), dtype=np.float32)

        spectrum_holder_view = spectrum_holder

        for index in range(neutral_mass_location.size):
            spectrum_holder_view[0, neutral_mass_location_view[index]] = max(
                spectrum_holder_view[0, neutral_mass_location_view[index]],
                norm_intensity_view[index])
        spectrum_original_forward = np.copy(spectrum_holder)
        spectrum_original_backward = np.copy(spectrum_holder)

        # peptide_mass
        spectrum_original_forward[0, int(round(
            peptide_mass * self.SPECTRUM_RESOLUTION))] = spectrum_intensity_max
        spectrum_original_backward[0, int(round(
            peptide_mass * self.SPECTRUM_RESOLUTION))] = spectrum_intensity_max
        # N-terminal, b-ion, peptide_mass_C
        # append N-terminal

        # append peptide_mass_C
        mass_C = self.mass_C_terminus + self.mass_H
        peptide_mass_C = peptide_mass - mass_C
        spectrum_original_forward[0, int(round(
            peptide_mass_C * self.SPECTRUM_RESOLUTION))] = spectrum_intensity_max
        # C-terminal, y-ion, peptide_mass_N

        # append peptide_mass_N
        mass_N = self.mass_N_terminus - self.mass_H
        peptide_mass_N = peptide_mass - mass_N
        spectrum_original_backward[0, int(round(
            peptide_mass_N * self.SPECTRUM_RESOLUTION))] = spectrum_intensity_max

        return spectrum_original_forward, spectrum_original_backward

    def get_location(self, peptide_mass, prefix_mass, direction):
        if direction == 0:
            candidate_b_mass = prefix_mass + self.mass_ID_np
            candidate_y_mass = peptide_mass - candidate_b_mass
        elif direction == 1:
            candidate_y_mass = prefix_mass + self.mass_ID_np
            candidate_b_mass = peptide_mass - candidate_y_mass

        # b-ions
        candidate_b_H2O = candidate_b_mass - self.mass_H2O
        candidate_b_NH3 = candidate_b_mass - self.mass_NH3
        candidate_b_plus2_charge1 = ((candidate_b_mass + 2 * self.mass_H) / 2
                                     - self.mass_H)

        # y-ions
        candidate_y_H2O = candidate_y_mass - self.mass_H2O
        candidate_y_NH3 = candidate_y_mass - self.mass_NH3
        candidate_y_plus2_charge1 = ((candidate_y_mass + 2 * self.mass_H) / 2
                                     - self.mass_H)
        # ion_8
        b_ions = [candidate_b_mass,
                  candidate_b_H2O,
                  candidate_b_NH3,
                  candidate_b_plus2_charge1]
        y_ions = [candidate_y_mass,
                  candidate_y_H2O,
                  candidate_y_NH3,
                  candidate_y_plus2_charge1]
        ion_mass_list = b_ions + y_ions
        ion_mass = np.array(ion_mass_list, dtype=np.float32)

        # ion locations
        location_sub50 = np.rint(ion_mass * self.SPECTRUM_RESOLUTION).astype(np.int32)
        location_sub50 -= (self.WINDOW_SIZE // 2)
        location_plus50 = location_sub50 + self.WINDOW_SIZE
        ion_id_rows, aa_id_cols = np.nonzero(np.logical_and(
            location_sub50 >= 0,
            location_plus50 <= self.MZ_SIZE))
        return ion_id_rows, aa_id_cols, location_sub50, location_plus50

    def copy_values(self, candidate_intensity_view, spectrum_view, location_sub, i1, i2):

        i1_start = self.neighbor_size * i1
        for neighbor in range(self.neighbor_size):
            for j in range(self.WINDOW_SIZE):
                try:
                    candidate_intensity_view[i2, i1_start + neighbor, j] = spectrum_view[
                        neighbor, location_sub[i1, i2] + j]
                except:
                    continue

    def get_candidate_intensity(self, spectrum_original, peptide_mass, prefix_mass, direction):
        ion_id_rows, aa_id_cols, location_sub50, location_plus50 = self.get_location(peptide_mass, prefix_mass,
                                                                                     direction)
        # candidate_intensity
        candidate_intensity = np.zeros(shape=(self.vocab_size,
                                              self.neighbor_size * self.num_ion,
                                              self.WINDOW_SIZE),
                                       dtype=np.float32)

        location_sub50_view = location_sub50

        location_plus50_view = location_plus50
        candidate_intensity_view = candidate_intensity
        row = ion_id_rows.astype(np.int32)
        col = aa_id_cols.astype(np.int32)

        for index in range(ion_id_rows.size):
            if col[index] < 3:
                continue
            self.copy_values(candidate_intensity_view, spectrum_original, location_sub50_view, row[index], col[index])
        max_intensity = np.max(candidate_intensity)
        if max_intensity > 1.0:
            candidate_intensity /= max_intensity
        return candidate_intensity

    def _parse_spectrum(self, precursor_mass, mz_lists, intensity_lists, neighbor_right_count, neighbor_size_half):

        MZ_SIZE = int(self.MZ_MAX * self.SPECTRUM_RESOLUTION)
        spectrum_original_forward_list = []
        spectrum_original_backward_list = []
        for mz_list, intensity_list in zip(mz_lists, intensity_lists):
            spectrum_original_forward, spectrum_original_backward = self.process_spectrum(mz_list,
                                                                                          intensity_list,
                                                                                          precursor_mass)
            spectrum_original_forward_list.append(spectrum_original_forward)
            spectrum_original_backward_list.append(spectrum_original_backward)

        if neighbor_right_count < neighbor_size_half:
            for x in range(neighbor_size_half - neighbor_right_count):
                spectrum_original_forward_list.append(np.zeros(
                    shape=(1, self.MZ_SIZE),
                    dtype=np.float32))
                spectrum_original_backward_list.append(np.zeros(
                    shape=(1, self.MZ_SIZE),
                    dtype=np.float32))

        spectrum_original_forward = np.vstack(spectrum_original_forward_list)
        spectrum_original_backward = np.vstack(spectrum_original_backward_list)

        return spectrum_original_forward, spectrum_original_backward

    def _process_raw_seq(self, raw_seq):
        peptide = []
        raw_len = len(raw_seq)
        pep_len, index = 0, 0
        while index < raw_len:
            if raw_seq[index] == '+':
                if peptide[-1] == 'C' and raw_seq[index:index + 7] == '+57.021':
                    peptide[-1] = 'C(Carbamidomethylation)'
                    index += 7
            else:
                peptide.append(raw_seq[index])
                index += 1
                pep_len += 1
        return peptide, pep_len

    def calculate_ms2(self, precursor_mz, precursor_charge, scan_list_middle, mz_list, int_list, neighbor_right_count,
                      neighbor_size_half, seq):
        candidate_intensity_list_forward = []
        candidate_intensity_list_backward = []

        precursor_mass = precursor_mz * precursor_charge - self.mass_H * precursor_charge

        if precursor_mass > self.MZ_MAX:
            return None
        spectrum_original_forward, spectrum_original_backward = self._parse_spectrum(precursor_mass, mz_list, int_list,
                                                                                     neighbor_right_count,
                                                                                     neighbor_size_half)
        peptide, peptide_len = self._process_raw_seq(seq)
        for bucket_id, target_size in enumerate(self._buckets):
            if peptide_len + 2 <= target_size:  # +2 to include GO and EOS
                break
        decoder_size = self._buckets[bucket_id]
        peptide_ids = [self.vocab[x] for x in peptide]
        pad_size = decoder_size - (len(peptide_ids) + 2)
        # forward

        peptide_ids_forward = peptide_ids[:]
        peptide_ids_forward.insert(0, self.GO_ID)
        peptide_ids_forward.append(self.EOS_ID)
        peptide_ids_forward += [self.PAD_ID] * pad_size

        peptide_ids_backward = peptide_ids[::-1]
        peptide_ids_backward.insert(0, self.EOS_ID)
        peptide_ids_backward.append(self.GO_ID)
        peptide_ids_backward += [self.PAD_ID] * pad_size

        prefix_mass = 0.0
        suffix_mass = 0.0
        for index in range(self._buckets[-1]):
            if index < decoder_size:
                prefix_mass += self.mass_ID[peptide_ids_forward[index]]
                candidate_intensity_forward = self.get_candidate_intensity(
                    spectrum_original_forward,
                    precursor_mass,
                    prefix_mass,
                    0)
                cand_int_shape_forward = candidate_intensity_forward.shape

                suffix_mass += self.mass_ID[peptide_ids_backward[index]]
                candidate_intensity_backward = self.get_candidate_intensity(
                    spectrum_original_backward,
                    precursor_mass,
                    suffix_mass,
                    1)
                cand_int_shape_backward = candidate_intensity_backward.shape
            else:
                candidate_intensity_forward = np.zeros(cand_int_shape_forward)
                candidate_intensity_backward = np.zeros(cand_int_shape_backward)

            candidate_intensity_list_forward.append(candidate_intensity_forward)
            candidate_intensity_list_backward.append(candidate_intensity_backward)
        return candidate_intensity_list_forward, candidate_intensity_list_backward



"""Parse mass spectra into an HDF5 file format."""
import logging
from pathlib import Path

import h5py
import numpy as np
import pickle
import bz2

from .. import utils
from .parsers import MzmlParser, MzxmlParser, MgfParser

from .dia_io import WorkerDIA


LOGGER = logging.getLogger(__name__)


class SpectrumIndex:
    """Store and access a collection of mass spectra.

    This class parses one or more mzML file, converts it to an HDF5 file
    format. This allows depthcharge to access spectra from many different
    files quickly and without loading them all into memory.

    Parameters
    ----------
    index_path : str
        The name and path of the HDF5 file index. If the path does
        not contain the `.h5` or `.hdf5` extension, `.hdf5` will be added.
    ms_data_files : str or list of str, optional
        The mzML to include in this collection.
    ms_level : int, optional
        The level of tandem mass spectra to use.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    annotated : bool, optional
        Whether or not the index contains spectrum annotations.
    overwite : bool, optional
        Overwrite previously indexed files? If ``False`` and new files are
        provided, they will be appended to the collection.

    Attributes
    ----------
    ms_files : list of str
    path : Path
    ms_level : int
    valid_charge : Optional[Iterable[int]]
    annotated : bool
    overwrite : bool
    n_spectra : int
    n_peaks : int
    """

    def __init__(
        self,
        index_path,
        ms_data_files=None,
        feature_file = None,
        ms_level=2,
        valid_charge=None,
        annotated=False,
        overwrite=False,
    ):
        """Initialize a SpectrumIndex"""
        index_path = Path(index_path)
        if index_path.suffix not in [".h5", ".hdf5"]:
            index_path = Path(str(index_path) + ".hdf5")

        # Set attributes and check parameters:
        self._path = index_path
        self._ms_level = utils.check_positive_int(ms_level, "ms_level")
        self._valid_charge = valid_charge
        self._annotated = bool(annotated)
        self._overwrite = bool(overwrite)
        self._handle = None
        self._file_offsets = np.array([0])
        self._file_map = {}
        self._locs = {}
        self.precursor_info = {}
        self.worker_dia_io = WorkerDIA()


        #read feature file
        if feature_file is not None:
            self.read_feature_file(feature_file)
        # Create the file if it doesn't exist.
        if not self.path.exists() or self.overwrite:
            with h5py.File(self.path, "w") as index:
                index.attrs["ms_level"] = self.ms_level
                index.attrs["n_spectra"] = 0
                index.attrs["n_peaks"] = 0
                index.attrs["annotated"] = self.annotated

        # Else, verify that the previous index uses the same ms_level.
        else:
            with self:
                try:
                    assert self._handle.attrs["ms_level"] == self.ms_level
                    assert self._handle.attrs["annotated"] == self.annotated
                except (KeyError, AssertionError):
                    raise ValueError(
                        f"'{self.path}' already exists, but was created with "
                        "incompatible parameters. Use 'overwrite=True' to "
                        "overwrite it."
                    )

                self._reindex()

        # Now parse spectra.
        if ms_data_files is not None:
            ms_data_files = utils.listify(ms_data_files)
            LOGGER.info("Reading %i files...", len(ms_data_files))
            for ms_file in ms_data_files:
                self.add_file(ms_file)

    def read_feature_file(self, feature_file):
        with open(feature_file, "rb") as pklfile:
            self.precursor_info = pickle.load(pklfile)


    def _reindex(self):
        """Update the file mappings and offsets."""
        offsets = []
        for idx in range(len(self._handle)):
            grp = self._handle[str(idx)]
            offsets.append(grp.attrs["n_spectra"])
            self._file_map[grp.attrs["path"]] = idx

        self._file_offsets = np.cumsum([0] + offsets)

        # Build a map of 1D indices to 2D locations:
        grp_idx = 0
        for lin_idx in range(offsets[-1]):
            grp_idx += lin_idx >= offsets[grp_idx]
            row_idx = lin_idx - offsets[grp_idx - 1]
            self._locs[lin_idx] = (grp_idx, row_idx)

    def _get_parser(self, ms_data_file):
        """Get the parser for the MS data file.

        Parameters
        ----------
        ms_data_file : Path
            The mass spectrometry data file to be parsed.

        Returns
        -------
        MzmlParser, MzxmlParser, or MgfParser
            The appropriate parser for the file.
        """
        kw_args = dict(ms_level=self.ms_level, valid_charge=self.valid_charge)

        if ms_data_file.suffix.lower() == ".mzml":
            return MzmlParser(ms_data_file, **kw_args)

        if ms_data_file.suffix.lower() == ".mzxml":
            return MzxmlParser(ms_data_file, **kw_args)

        if ms_data_file.suffix.lower() == ".mgf":
            return MgfParser(ms_data_file, **kw_args)

        raise ValueError("Only mzML, mzXML, and MGF files are supported.")

    def _assemble_metadata(self, parser):
        """Assemble the metadata.

        Parameters
        ----------
        parser : MzmlParser or MgfParser
            The parser to use.

        Returns
        -------
        numpy.ndarray of shape (n_spectra,)
            The file metadata.
        """
        meta_types = [
            ("precursor_mz", np.float32),
            ("precursor_charge", np.uint8),
            ("offset", np.uint64),
            ("scan_id", np.uint32),
        ]

        metadata = np.empty(parser.n_spectra, dtype=meta_types)
        metadata["precursor_mz"] = parser.precursor_mz
        metadata["precursor_charge"] = parser.precursor_charge
        metadata["offset"] = parser.offset
        metadata["scan_id"] = parser.scan_id
        return metadata

    def add_file(self, ms_data_file):
        """Add a mass spectrometry data file to the index.

        Parameters
        ----------
        ms_data_file : str or Path
            The mass spectrometry data file to add. It must be in an mzML or
            MGF file format and use an ``.mzML``, ``.mzXML``, or ``.mgf`` file extension.
        """
        ms_data_file = Path(ms_data_file)
        if str(ms_data_file) in self._file_map:
            return

        # Read the file:
        parser = self._get_parser(ms_data_file)
        parser.read()

        # Create the tables:
        metadata = self._assemble_metadata(parser)

        spectrum_types = [
            ("mz_array", np.float64),
            ("intensity_array", np.float32),
        ]

        spectra = np.zeros(parser.n_peaks, dtype=spectrum_types)
        spectra["mz_array"] = parser.mz_arrays
        spectra["intensity_array"] = parser.intensity_arrays

        # Write the tables:
        with h5py.File(self.path, "a") as index:
            group_index = len(index)
            group = index.create_group(str(group_index))
            group.attrs["path"] = str(ms_data_file)
            group.attrs["n_spectra"] = parser.n_spectra
            group.attrs["n_peaks"] = parser.n_peaks
            group.attrs["id_type"] = parser.id_type

            # Update overall stats:
            index.attrs["n_spectra"] += parser.n_spectra
            index.attrs["n_peaks"] += parser.n_peaks

            # Add the datasets:
            group.create_dataset(
                "metadata",
                data=metadata,
            )

            group.create_dataset(
                "spectra",
                data=spectra,
            )
            
            group.create_dataset(
               "scan",
               data=parser.scans,
               dtype=h5py.string_dtype(),
            )
            try:
                group.create_dataset(
                    "annotations",
                    data=parser.annotations,
                    dtype=h5py.string_dtype(),
                )
            except (KeyError, AttributeError):
                pass

            self._file_map[str(ms_data_file)] = group_index
            end_offset = self._file_offsets[-1] + parser.n_spectra
            self._file_offsets = np.append(self._file_offsets, [end_offset])

            # Update the locations:
            grp_idx = len(self._file_offsets) - 2
            for row_idx in range(parser.n_spectra):
                lin_idx = row_idx + self._file_offsets[-2]
                self._locs[lin_idx] = (grp_idx, row_idx)

    def get_ms2_spectra(self, spec_info, scan_id):

        for pep_seq in self.precursor_info.keys():
          if len(spec_info) > 4:
              break
          for precursor in self.precursor_info[pep_seq]:
            if spec_info[2] == np.float32(precursor['precursor_mz']) and spec_info[3] == precursor['precursor_charge'] and scan_id in precursor["scan_list_middle"]:
                ms1 = precursor['ms1']
                ms2_b_ion, ms2_y_ion = self.worker_dia_io.calculate_ms2(precursor['precursor_mz'],
                                                                      precursor['precursor_charge'],
                                                                      precursor["scan_list_middle"], \
                                                                      precursor["mz_list"], precursor["int_list"],
                                                                      precursor["neighbor_right_count"],
                                                                      precursor["neighbor_size_half"], pep_seq)
                spec_info = spec_info + (ms1,)
                spec_info = spec_info + (ms2_b_ion,)
                spec_info = spec_info + (ms2_y_ion,)
                break
        
        return spec_info

    def get_spectrum(self, idx):
        """Access a mass spectrum.

        Parameters
        ----------
        idx : int
            The index of the index of the mass spectrum to look-up.

        Returns
        -------
        tuple of numpy.ndarray
            The m/z values, intensity values, precurosr m/z, precurosr charge,
            and the spectrum annotation.
        """
        group_index, row_index = self._locs[idx]
        if self._handle is None:
            raise RuntimeError("Use the context manager for access.")

        grp = self._handle[str(group_index)]
        metadata = grp["metadata"]
        spectra = grp["spectra"]
        annotations = grp["annotations"]
        offsets = metadata["offset"][row_index : row_index + 2]
        scan = grp["scan"][row_index].decode()
        start_offset = offsets[0]
        if offsets.shape[0] == 2:
            stop_offset = offsets[1]
        else:
            stop_offset = spectra.shape[0]

        spectrum = spectra[start_offset:stop_offset]
        precursor = metadata[row_index]
        out = (
            spectrum["mz_array"],
            spectrum["intensity_array"],
            precursor["precursor_mz"],
            precursor["precursor_charge"],
        
            )
        if annotations.shape:
            return out, scan

        out = self.get_ms2_spectra(out, scan)

        return out

    def get_spectrum_id(self, idx):
        """Get the identifier for a mass spectrum.

        Parameters
        ----------
        idx : int
            The index of the mass spectrum in the SpectrumIndex.

        Returns
        -------
        ms_data_file : str
            The mass spectrometry data file from which the mass spectrum was
            originally parsed.
        identifier : str
            The mass spectrum identifier, per PSI recommendations.
        """
        group_index, row_index = self._locs[idx]
        if self._handle is None:
            raise RuntimeError("Use the context manager for access.")

        grp = self._handle[str(group_index)]
        ms_data_file = grp.attrs["path"]
        identifier = grp["metadata"][row_index]["scan_id"]
        prefix = grp.attrs["id_type"]
        return ms_data_file, f"{prefix}={identifier}"

    def __len__(self):
        """The number of spectra in the index."""
        return self._offsets[-1]

    def __getitem__(self, idx):
        """Access a mass spectrum.

        Parameters
        ----------
        idx : int
            The overall index of the mass spectrum to retrieve.

        Returns
        -------
        tuple of numpy.ndarray
            The m/z values, intensity values, precurosr m/z, precurosr charge,
            and the annotation (if available).
        """
        if self._handle is None:
            with self:
                return self.get_spectrum(idx)

        return self.get_spectrum(idx)

    def __enter__(self):
        """Open the index file for reading."""
        self._handle = h5py.File(
            self.path,
            "r",
            rdcc_nbytes=int(3e8),
            rdcc_nslots=1024000,
        )
        return self

    def __exit__(self, *args):
        """Close the HDF5 file."""
        self._handle.close()
        self._handle = None

    @property
    def ms_files(self):
        """The files currently in the index."""
        return list(self._file_map.keys())

    @property
    def path(self):
        """The path to the underyling HDF5 index file."""
        return self._path

    @property
    def ms_level(self):
        """The MS level of tandem mass spectra in the collection."""
        return self._ms_level

    @property
    def valid_charge(self):
        """Valid precursor charges for spectra to be included."""
        return self._valid_charge

    @property
    def annotated(self):
        """Whether or not the index contains spectrum annotations."""
        return self._annotated

    @property
    def overwrite(self):
        """Overwrite a previous index?"""
        return self._overwrite

    @property
    def n_spectra(self):
        """The total number of mass spectra in the index."""
        if self._handle is None:
            with self:
                return self._handle.attrs["n_spectra"]

        return self._handle.attrs["n_spectra"]

    @property
    def n_peaks(self):
        """The total number of mass peaks in the index."""
        if self._handle is None:
            with self:
                return self._handle.attrs["n_peaks"]

        return self._handle.attrs["n_peaks"]


class AnnotatedSpectrumIndex(SpectrumIndex):
    """Store and access a collection of annotated mass spectra.

    This class parses one or more mzML file, converts it to our HDF5 index
    format. This allows us to access spectra from many different files quickly
    and without loading them all into memory.

    Parameters
    ----------
    index_path : str
        The name and path of the HDF5 file index. If the path does
        not contain the `.h5` or `.hdf5` extension, `.hdf5` will be added.
    ms_data_files : str or list of str, optional
        The MGF to include in this collection.
    ms_level : int, optional
        The level of tandem mass spectra to use.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    overwite : bool
        Overwrite previously indexed files? If ``False`` and new files are
        provided, they will be appended to the collection.

    Attributes
    ----------
    ms_files : list of str
    path : Path
    ms_level : int
    valid_charge : Optional[Iterable[int]]
    overwrite : bool
    n_spectra : int
    n_peaks : int
    """

    def __init__(
        self,
        index_path,
        ms_data_files=None,
        feature_file = None,
        ms_level=2,
        valid_charge=None,
        overwrite=False,
    ):
        """Initialize an AnnotatedSpectrumIndex"""
        super().__init__(
            index_path=index_path,
            ms_data_files=ms_data_files,
            feature_file = feature_file,
            ms_level=ms_level,
            valid_charge=valid_charge,
            annotated=True,
            overwrite=overwrite,
        )

    def _get_parser(self, ms_data_file):
        """Get the parser for the MS data file"""
        if ms_data_file.suffix.lower() == ".mgf":
            return MgfParser(
                ms_data_file,
                ms_level=self.ms_level,
                annotations=True,
            )

        raise ValueError("Only MGF files are currently supported.")

    def get_spectrum(self, idx):
        """Access a mass spectrum.

        Parameters
        ----------
        idx : int
            The index of the mass spectrum in the AnnotatedSpectrumIndex.

        Returns
        -------
        tuple of numpy.ndarray
            The m/z values, intensity values, precurosr m/z, precurosr charge,
            and the spectrum annotation.
        """
        spec_info, scan_id = super().get_spectrum(idx)

        group_index, row_index = self._locs[idx]
        grp = self._handle[str(group_index)]
        annotations = grp["annotations"]
        spec_ann = annotations[row_index].decode()
        if self.precursor_info is None:
            print('self.precursor_info is empty')
            return (*spec_info, spec_ann)

        for precursor in self.precursor_info[spec_ann]:

           if spec_info[2] == np.float32(precursor['precursor_mz']) and spec_info[3] == precursor['precursor_charge'] and scan_id in precursor["scan_list_middle"]:
                ms1 = precursor['ms1']
                ms2_b_ion, ms2_y_ion = self.worker_dia_io.calculate_ms2(precursor['precursor_mz'], precursor['precursor_charge'],
                        precursor["scan_list_middle"], precursor["mz_list"], precursor["int_list"],
                        precursor["neighbor_right_count"], precursor["neighbor_size_half"],
                        spec_ann)
                spec_info = spec_info + (ms1,)
                spec_info = spec_info + (ms2_b_ion,)
                spec_info = spec_info +(ms2_y_ion, )

                break

        
        return (*spec_info, spec_ann)



"""Similarity measures between mass spectra."""
import numba as nb
import numpy as np


@nb.njit
def gsp(spectrum_x, spectrum_y, tol):
    """Compute the dot product between two spectra.

    This code was adapted from Wout:
    https://github.com/bittremieux/GLEAMS/blob/master/gleams/feature/spectrum.py#L154


    Parameters
    ----------
    spectrum_x : np.ndarray
    spectrum_y : np.ndarray
        The spectra to compare. Each row should be a peak, with the first
        column indicating the m/z and the second column indicating the
        intensities.
    tol : float
        The fragment m/z tolerance used to match peaks in both spectra with
        each other.

    Returns
    -------
    float
        The dot product between both spectra.
    """
    # Find the matching peaks between both spectra.
    peak_match_scores, peak_match_idx = [], []
    peak_other_i = 0
    mz_other, intensity_other = spectrum_y.T
    for peak_i, (peak_mz, peak_intensity) in enumerate(spectrum_x):
        # Advance while there is an excessive mass difference.
        while (
            peak_other_i < len(mz_other) - 1
            and peak_mz - tol > mz_other[peak_other_i]
        ):
            peak_other_i += 1
        # Match the peaks within the fragment mass window if possible.
        peak_other_window_i = peak_other_i
        while (
            peak_other_window_i < len(mz_other)
            and abs(peak_mz - (mz_other[peak_other_window_i])) <= tol
        ):
            peak_match_scores.append(
                peak_intensity * intensity_other[peak_other_window_i]
            )
            peak_match_idx.append((peak_i, peak_other_window_i))
            peak_other_window_i += 1

    score = 0
    if len(peak_match_scores) > 0:
        # Use the most prominent peak matches to compute the score (sort in
        # descending order).
        peak_match_scores_arr = np.asarray(peak_match_scores)
        peak_match_order = np.argsort(peak_match_scores_arr)[::-1]
        peak_match_scores_arr = peak_match_scores_arr[peak_match_order]
        peak_match_idx_arr = np.asarray(peak_match_idx)[peak_match_order]
        peaks_used, peaks_used_other = set(), set()
        for peak_match_score, peak_i, peak_other_i in zip(
            peak_match_scores_arr,
            peak_match_idx_arr[:, 0],
            peak_match_idx_arr[:, 1],
        ):
            if (
                peak_i not in peaks_used
                and peak_other_i not in peaks_used_other
            ):
                score += peak_match_score
                # Make sure these peaks are not used anymore.
                peaks_used.add(peak_i)
                peaks_used_other.add(peak_other_i)

    return score



"""Simple encoders for input into Transformers and the like."""
import torch
import einops
import numpy as np


class MassEncoder(torch.nn.Module):
    """Encode mass values using sine and cosine waves.

    Parameters
    ----------
    dim_model : int
        The number of features to output.
    min_wavelength : float
        The minimum wavelength to use.
    max_wavelength : float
        The maximum wavelength to use.
    """

    def __init__(self, dim_model, min_wavelength=0.001, max_wavelength=10000):
        """Initialize the MassEncoder"""
        super().__init__()

        n_sin = int(dim_model / 2)
        n_cos = dim_model - n_sin

        if min_wavelength:
            base = min_wavelength / (2 * np.pi)
            scale = max_wavelength / min_wavelength
        else:
            base = 1
            scale = max_wavelength / (2 * np.pi)

        sin_term = base * scale ** (
            torch.arange(0, n_sin).float() / (n_sin - 1)
        )
        cos_term = base * scale ** (
            torch.arange(0, n_cos).float() / (n_cos - 1)
        )

        self.register_buffer("sin_term", sin_term)
        self.register_buffer("cos_term", cos_term)

    def forward(self, X):
        """Encode m/z values.

        Parameters
        ----------
        X : torch.Tensor of shape (n_masses)
            The masses to embed.

        Returns
        -------
        torch.Tensor of shape (n_masses, dim_model)
            The encoded features for the mass spectra.
        """
        sin_mz = torch.sin(X / self.sin_term)
        cos_mz = torch.cos(X / self.cos_term)
        return torch.cat([sin_mz, cos_mz], axis=-1)


class PeakEncoder(MassEncoder):
    """Encode m/z values in a mass spectrum using sine and cosine waves.

    Parameters
    ----------
    dim_model : int
        The number of features to output.
    dim_intensity : int, optional
        The number of features to use for intensity. The remaining features
        will be used to encode the m/z values.
    min_wavelength : float, optional
        The minimum wavelength to use.
    max_wavelength : float, optional
        The maximum wavelength to use.
    """

    def __init__(
        self,
        dim_model,
        dim_intensity=None,
        min_wavelength=0.001,
        max_wavelength=10000,
    ):
        """Initialize the MzEncoder"""
        self.dim_intensity = dim_intensity
        self.dim_mz = dim_model
        if self.dim_intensity is not None:
            self.dim_mz -= self.dim_intensity

        super().__init__(
            dim_model=self.dim_mz,
            min_wavelength=min_wavelength,
            max_wavelength=max_wavelength,
        )

        if self.dim_intensity is None:
            self.int_encoder = torch.nn.Linear(1, dim_model, bias=False)
        else:
            self.int_encoder = MassEncoder(
                dim_model=dim_intensity,
                min_wavelength=0,
                max_wavelength=1,
            )

    def forward(self, X):
        """Encode m/z values and intensities.

        Note that we expect intensities to fall within the interval [0, 1].

        Parameters
        ----------
        X : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra to embed. Axis 0 represents a mass spectrum, axis 1
            contains the peaks in the mass spectrum, and axis 2 is essentially
            a 2-tuple specifying the m/z-intensity pair for each peak. These
            should be zero-padded, such that all of the spectra in the batch
            are the same length.

        Returns
        -------
        torch.Tensor of shape (n_spectr, n_peaks, dim_model)
            The encoded features for the mass spectra.
        """
        m_over_z = X[:, :, [0]]
        encoded = super().forward(m_over_z)
        intensity = self.int_encoder(X[:, :, [1]])
        if self.dim_intensity is None:
            return encoded + intensity

        return torch.cat([encoded, intensity], dim=2)


class PositionalEncoder(torch.nn.Module):
    """The positional encoder for sequences.

    Parameters
    ----------
    dim_model : int
        The number of features to output.
    """

    def __init__(self, dim_model, max_wavelength=10000):
        """Initialize the MzEncoder"""
        super().__init__()

        n_sin = int(dim_model / 2)
        n_cos = dim_model - n_sin
        scale = max_wavelength / (2 * np.pi)

        sin_term = scale ** (torch.arange(0, n_sin).float() / (n_sin - 1))
        cos_term = scale ** (torch.arange(0, n_cos).float() / (n_cos - 1))
        self.register_buffer("sin_term", sin_term)
        self.register_buffer("cos_term", cos_term)

    def forward(self, X):
        """Encode positions in a sequence.

        Parameters
        ----------
        X : torch.Tensor of shape (batch_size, n_sequence, n_features)
            The first dimension should be the batch size (i.e. each is one
            peptide) and the second dimension should be the sequence (i.e.
            each should be an amino acid representation).

        Returns
        -------
        torch.Tensor of shape (batch_size, n_sequence, n_features)
            The encoded features for the mass spectra.
        """
        pos = torch.arange(X.shape[1]).type_as(self.sin_term)
        pos = einops.repeat(pos, "n -> b n", b=X.shape[0])
        sin_in = einops.repeat(pos, "b n -> b n f", f=len(self.sin_term))
        cos_in = einops.repeat(pos, "b n -> b n f", f=len(self.cos_term))

        sin_pos = torch.sin(sin_in / self.sin_term)
        cos_pos = torch.cos(cos_in / self.cos_term)
        encoded = torch.cat([sin_pos, cos_pos], axis=2)
        return encoded + X
